#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import os
import sys
import shutil
from itertools import chain
import inspect

import pkg_resources
import yaml

try:
    from cloudfoundry import contexts
except ImportError:
    sys.path.append('.')
    from cloudfoundry import contexts
from cloudfoundry import utils
from cloudfoundry.path import path


class CharmGenerator(object):
    author = "CloudFoundry Charm Generator <cs:~cf-charmers/cloudfoundry>"

    def __init__(self, releases, service_registry, from_charmstore=True,
                 cs_namespace='~cf-charmers', admin_ui=True):
        self.__releases = releases
        self.release = None
        self.release_version = None
        self.service_registry = service_registry
        self.from_charmstore = from_charmstore
        self.cs_namespace = cs_namespace
        self.admin_ui = admin_ui

    def select_release(self, version):
        if isinstance(version, basestring):
            try:
                version = int(version)
            except ValueError:
                raise KeyError(version)

        for r in self.__releases:
            versions = r['releases']
            low = versions[0]
            high = None
            if len(versions) == 2:
                high = versions[1]
            if version < low:
                continue
            if not high or version <= high:
                self.release = r
                self.release_version = version
                return r
        raise KeyError(version)

    def _is_relation(self, context):
        return inspect.isclass(context)\
            and issubclass(context, contexts.RelationContext)

    def build_metadata(self, service_key):
        # service usage within the topo can include the service name
        # allowing this to be a tuple
        if isinstance(service_key, (tuple, list)):
            service_key = service_key[0]
        service = self.service_registry[service_key]
        result = dict(
            name=service_key,
            summary=service.get('summary', ''),
            description=service.get('description', ''),
            author=self.author,
            requires={
                contexts.OrchestratorRelation.name: dict(
                    interface=contexts.OrchestratorRelation.interface)
            })
        provides = {}
        for job in service.get('jobs', []):
            for relation in job.get('provided_data', []):
                if self._is_relation(relation):
                    provides[relation.name] = dict(
                        interface=relation.interface)
            for relation in job.get('required_data', []):
                if self._is_relation(relation):
                    result['requires'][relation.name] = dict(
                        interface=relation.interface)
        if provides:
            result['provides'] = provides

        return result

    def build_hooks(self, service_key):
        meta = self.build_metadata(service_key)
        results = ['start', 'stop', 'config-changed',
                   'upgrade-charm', 'install', 'health']
        for rel in chain(meta.get('provides', {}), meta.get('requires', {})):
            results.append('{}-relation-changed'.format(rel))
            results.append('{}-relation-joined'.format(rel))
            results.append('{}-relation-broken'.format(rel))
        return results

    def build_entry(self, service_key):
        _, name, _ = self._parse_charm_ref(service_key)
        return "\n".join([
            '#!/usr/bin/env python2.7',
            'from cloudfoundry.jobs import job_manager',
            'job_manager("{}")'.format(name)
        ])

    def generate_charm(self, service_key, target_dir):
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        meta = self.build_metadata(service_key)
        meta_target = open(os.path.join(target_dir, 'metadata.yaml'), 'w')
        yaml.safe_dump(meta, meta_target, default_flow_style=False)
        meta_target.close()

        shutil.copyfile(pkg_resources.resource_filename(
            __name__, '../icon.svg'),
            os.path.join(target_dir, 'icon.svg'))

        hook_dir = os.path.join(target_dir, 'hooks')
        if not os.path.exists(hook_dir):
            os.makedirs(hook_dir)
        entry = os.path.join(target_dir, 'hooks', 'entry.py')

        with open(entry, 'w') as target:
            os.fchmod(target.fileno(), 0755)
            target.write(self.build_entry(service_key))

        for hook in self.build_hooks(service_key):
            os.symlink('entry.py', os.path.join(hook_dir, hook))

    def _build_charm_ref(self, charm_id):
        if charm_id.startswith('cs:'):
            return dict(charm=charm_id)
        else:
            if self.from_charmstore:
                namespace = '{}/'.format(self.cs_namespace) if self.cs_namespace else ''
                return {"charm": "cs:{}trusty/{}".format(namespace, charm_id)}
            else:
                return {
                    "charm": charm_id,
                    "branch": "local:trusty/{}".format(charm_id)
                }

    def _parse_charm_ref(self, service_id):
        if isinstance(service_id, tuple):
            charm_id = charm_name = service_id[0]
            service_name = service_id[1]
        else:
            charm_id = charm_name = service_id
            service_name = service_id

        if '/' in charm_name:
            charm_name = charm_name.split('/', 1)[1]

        if '/' in service_name:
            service_name = service_name.split('/', 1)[1]
        return charm_id, charm_name, service_name

    def _normalize_relation(self, rel):
        if isinstance(rel, tuple):
            return "{}:{}".format(*rel)
        else:
            return rel

    def _get_managed_charms(self):
        charms = []
        for service in self.release['topology']['services']:
            charm_id, charm_name, service_name = self._parse_charm_ref(service)
            if charm_id.startswith('cs:'):
                continue
            if charm_name not in self.service_registry:
                raise KeyError(
                    'Missing service_registry definition for charm: {}'.format(
                        charm_name))
            charms.append((charm_id, charm_name, service_name))
        return charms

    def _get_relations(self):
        relations = self.release['topology'].get('relations', [])
        services = {service_name: self.service_registry[charm_name]
                    for _, charm_name, service_name in self._get_managed_charms()}
        provided = {provider.name: service_name
                    for service_name, service_def in services.iteritems()
                    for job in service_def.get('jobs', [])
                    for provider in job.get('provided_data', [])}
        for service_name, service_def in services.iteritems():
            for job in service_def.get('jobs', []):
                for required in job.get('required_data', []):
                    if not self._is_relation(required):
                        continue
                    if required.name not in provided:
                        continue
                    provider_name = provided[required.name]
                    lhs = (service_name, required.name)
                    rhs = (provider_name, required.name)
                    relations.append((lhs, rhs))
                relations.append((
                    (service_name, contexts.OrchestratorRelation.name),
                    ('cloudfoundry', contexts.OrchestratorRelation.name),  # FIXME: don't hard-code service name
                ))
                if contexts.CloudFoundryCredentials in job.get('required_data'):
                    relations.append((
                        (service_name, contexts.CloudFoundryCredentials.name),
                        ('cloudfoundry', contexts.CloudFoundryCredentials.name),
                    ))
        return relations

    def load_placement_policy(self, policy, placements_file=None):
        if policy.startswith('manual:'):
            return yaml.safe_load(policy[7:])
        if placements_file is None:
            placements_file = pkg_resources.resource_filename(__name__, 'placements.yaml')
        with open(placements_file) as fp:
            policies = yaml.safe_load(fp)
        if policy not in policies:
            raise KeyError('Placement policy {} not found in {}'.format(policy, policies.keys()))
        return policies[policy]

    def _process_zones(self, services, zones, zone_default_directives=None):
        with_references = set()
        default_driver = False
        for zone in zones.values():
            if '*' in zone:
                zone.remove('*')
                default_driver = zone[0] if zone else None
            if zone:
                driver = zone[0]
                for follower in zone[1:]:
                    with_references.add(driver)
                    if 'to' not in services[follower]:
                        services[follower]['to'] = [driver]
        if default_driver is not False:
            for service_name, service in services.iteritems():
                if 'to' in service or service_name in with_references or service_name == default_driver:
                    continue
                if default_driver is None:
                    default_driver = service_name
                    continue
                service['to'] = [default_driver]
            if zone_default_directives:
                utils.deepmerge(services[default_driver], zone_default_directives)

    def _do_placement(self, services, placement):
        zones = placement.pop('zones', {})
        zone_default_directives = placement.pop('__default__', {})
        utils.deepmerge(services, placement)
        if zones:
            self._process_zones(services, zones, zone_default_directives)

    def build_deployment(self, placement=None):
        services = {'cloudfoundry': self._build_charm_ref('cloudfoundry')}
        services['cloudfoundry']['expose'] = True
        relations = []
        result = {'cloudfoundry': {
            # Trusty is Magic!
            'series': 'trusty',
            'services': services,
            'relations': relations
        }}

        rel_data = {}
        exposed = set(self.release['topology'].get('expose', []))
        constraints = self.release['topology'].get('constraints', {})
        for service_id in self.release['topology']['services']:
            charm_id, _, service_name = self._parse_charm_ref(service_id)
            services[service_name] = self._build_charm_ref(charm_id)
            if service_name in exposed:
                services[service_name]['expose'] = True
            constraint = constraints.get(service_name,
                                         constraints.get("__default__"))
            if constraint:
                services[service_name]['constraints'] = constraint

        for rel in self._get_relations():
            lhs = self._normalize_relation(rel[0])
            rhs = self._normalize_relation(rel[1])
            rel_data.setdefault(lhs, set()).add(rhs)
        for k, v in rel_data.items():
            relations.append((k, tuple(v)))

        if self.admin_ui:
            namespace = '{}/'.format(self.cs_namespace) if self.cs_namespace else ''
            services['webadmin'] = {
                'charm': 'cs:{}trusty/{}'.format(namespace, 'cf-webadmin'),
                'expose': True,
            }
            relations.append(('webadmin:db', ('mysql:db',)))
            relations.append(('webadmin:uaa', ('uaa:uaa',)))
            relations.append(('webadmin:uaadb', ('uaa:uaadb',)))
            relations.append(('webadmin:ccdb', ('cc:ccdb',)))
            relations.append(('webadmin:nats', ('nats:nats',)))
            relations.append(('webadmin:orchestrator', ('cloudfoundry:orchestrator',)))

        if placement:
            self._do_placement(services, placement)
        return result

    def generate_deployment(self, target_dir, placement=None):
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        bundle = self.build_deployment(placement)
        target = os.path.join(target_dir, 'bundles.yaml')
        with open(target, 'w') as fp:
            yaml.safe_dump(bundle, fp, default_flow_style=False)
            fp.flush()

    def generate(self, target_dir, placement=None):
        libroot = path(pkg_resources.resource_filename(__name__, '..'))
        cloudfoundry = libroot / 'cloudfoundry'
        charmhelpers = libroot / 'charmhelpers'
        files = libroot / 'files'
        wheelhouse = libroot / 'wheelhouse'

        repo = path(target_dir) / 'trusty'
        repo.makedirs_p()
        self.generate_deployment(target_dir, placement)

        if not self.from_charmstore:
            for _, charm_name, _ in self._get_managed_charms():
                charm_path = repo / charm_name
                charm_path.makedirs_p()

                if (charm_path / 'metadata.yaml').exists():
                    continue

                self.generate_charm(charm_name, charm_path)

                # vendor in libs and files
                cloudfoundry.copytree(charm_path / 'hooks' / 'cloudfoundry')
                charmhelpers.copytree(charm_path / 'hooks' / 'charmhelpers')
                files.copytree(charm_path / 'files')
                charm_wheelhouse = path(charm_path) / 'wheelhouse'
                charm_wheelhouse.makedirs_p()
                wheelhouse.files('raindance*.whl')[0].copy2(charm_wheelhouse)


def main(args=None):
    from cloudfoundry.releases import RELEASES
    from cloudfoundry.services import SERVICES
    parser = argparse.ArgumentParser()
    parser.add_argument('release', type=int)
    parser.add_argument('-A', '--no-admin-ui', dest="admin_ui", action="store_false")
    parser.add_argument('-d', '--directory', dest="directory")
    parser.add_argument('-f', '--force', action="store_true")
    parser.add_argument('-p', '--placement', default="sparse")
    parser.add_argument('-P', '--placements_file', default=None)
    parser.add_argument('-c', '--charmstore', action="store_true")
    options = parser.parse_args(args)

    using_default_dir = False
    if not options.directory:
        options.directory = "cloudfoundry-r{}".format(options.release)
        using_default_dir = True
    if not os.path.exists(options.directory):
        os.mkdir(options.directory)
    else:
        if options.force:
            shutil.rmtree(options.directory)
            os.mkdir(options.directory)
        elif using_default_dir is True:
            raise SystemExit("Release already generated: {}".format(
                options.directory))

    g = CharmGenerator(RELEASES, SERVICES, from_charmstore=options.charmstore,
                       admin_ui=options.admin_ui)
    g.select_release(options.release)
    placement = g.load_placement_policy(options.placement, options.placements_file)
    g.generate(options.directory, placement)

if __name__ == '__main__':
    main()
