import copy
import collections
import json
import logging
import os
import re
import socket
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager
from functools import partial, wraps

import requests
import yaml

from charmhelpers import fetch


ip_pat = re.compile('^(\d{1,3}\.){3}\d{1,3}$')
_client = None  # cached client connection


@contextmanager
def cd(directory):
    cwd = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(cwd)


def api_address():
    api_addresses = os.environ.get('JUJU_API_ADDRESSES')
    if api_addresses:
        return api_addresses.split()[0]
    return None


def record_pid():
    pid_dir = os.path.expanduser('~/.config/juju-deployer')
    if not os.path.exists(pid_dir):
        os.makedirs(pid_dir)
    with open(os.path.join(pid_dir, 'server.pid'), 'w') as fp:
        fp.write('%s\n' % os.getpid())


def flatten_relations(rels):
    result = []
    for pair in rels:
        a = pair[0]
        b = pair[1]
        if isinstance(b, list):
            for ep in b:
                result.append(tuple(sorted((a, ep))))
        else:
            result.append(tuple(sorted((a, b))))
    return set(result)


def flatten_reality(real):
    result = set()
    for s, d in real['Services'].items():
        rels = d .get('Relations', {})
        for iface, reps in rels.items():
            for rep in reps:
                result.add(tuple(
                    sorted(('%s:%s' % (s, iface), rep))))
    return result


def rel_exists(reality, end_a, end_b):
    # Checks for a named relation on one side that matches the local
    # endpoint and remote service.
    (name_a, name_b, rem_a, rem_b) = (end_a, end_b, None, None)

    if ":" in end_a:
        name_a, rem_a = end_a.split(":", 1)
    if ":" in end_b:
        name_b, rem_b = end_b.split(":", 1)

    try:
        rels_svc_a = reality['Services'][name_a].get('Relations', {})

        found = False
        for r, related in rels_svc_a.items():
            if name_b in related:
                if rem_a and r not in rem_a:
                    continue
                found = True
                break
        if found:
            return True
    except KeyError:
        pass
    return False


def deepmerge(dest, src):
    """
    Deep merge of two dicts.

    This is destructive (`dest` is modified), but values
    from `src` are passed through `copy.deepcopy`.
    """
    for k, v in src.iteritems():
        if dest.get(k) and isinstance(v, dict):
            deepmerge(dest[k], v)
        else:
            dest[k] = copy.deepcopy(v)
    return dest


class NestedDict(dict):
    def __init__(self, dict_or_iterable=None, **kwargs):
        if dict_or_iterable:
            if isinstance(dict_or_iterable, dict):
                self.update(dict_or_iterable)
            elif isinstance(dict_or_iterable, collections.Iterable):
                for k, v in dict_or_iterable:
                    self[k] = v
        if kwargs:
            self.update(kwargs)

    def __setitem__(self, key, value):
        key = key.split('.')
        o = self
        for part in key[:-1]:
            o = o.setdefault(part, self.__class__())
        dict.__setitem__(o, key[-1], value)

    def __getitem__(self, key):
        o = self
        if '.' in key:
            parts = key.split('.')
            key = parts[-1]
            for part in parts[:-1]:
                o = o[part]

        return dict.__getitem__(o, key)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def update(self, other):
        deepmerge(self, other)


def parse_config(conf_fn, defaults=None):
    with open(conf_fn, 'r') as fp:
        conf = NestedDict()
        if defaults:
            conf.update(defaults)
        conf.update(json.load(fp))
        return conf


def command(*base_args, **kwargs):
    check = kwargs.pop('check', False)
    throw = kwargs.pop('throw', True)

    def callable_command(*args, **kws):
        kwargs.update(kws)
        all_args = base_args + args
        logging.debug(' '.join(all_args))
        if 'env' not in kwargs:
            kwargs['env'] = os.environ
        logging.debug("invoke: %s", all_args)

        p = subprocess.Popen(all_args,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             **kwargs)
        output, _ = p.communicate()
        ret_code = p.poll()
        logging.debug('output: %s', output)
        if check is True:
            if ret_code != 0 and throw:
                raise subprocess.CalledProcessError(ret_code, all_args, output=output)
            return ret_code
        else:
            return output.strip()
    return callable_command


class Commander(object):
    def __getattr__(self, key):
        return command(key)

    def check(self, *args, **kwargs):
        return command(*args, check=True, **kwargs)

    def __call__(self, cmd, **kwargs):
        subprocess.check_output(cmd, shell=True, **kwargs)


sh = Commander()
dig = command('dig', '+short')
api_endpoints = sh.check('juju', 'api-endpoints', throw=False)


def current_env():
    return sh.juju('switch')


def get_jenv():
    cenv = current_env()
    return os.path.join(os.environ.get('JUJU_HOME',
                                       os.path.expanduser('~/.juju')),
                        'environments/%s.jenv' % cenv)


def wait_for(timeout, interval, *callbacks, **kwargs):
    """
    Repeatedly try callbacks until all return True

    This will wait interval seconds between attempts and will error out
    after timeout has been exceeded.

    Callbacks will be called with the container as their argument.

    Setting timeout to zero will loop until cancelled, power runs outs,
    hardware fails, or the heat death of the universe.
    """
    start = time.time()
    if timeout:
        end = start + timeout
    else:
        end = 0

    bar = kwargs.get('bar', None)
    message = kwargs.get('message', None)
    once = 1
    while True:
        passes = True
        if end > 0 and time.time() > end:
            raise OSError("Timeout exceeded in wait_for")
        if bar:
            bar.next(once, message=message)
            if once == 1:
                once = 0
        if int(time.time()) % interval == 0:
            for callback in callbacks:
                result = callback()
                passes = passes & bool(result)
                if passes is False:
                    break
            if passes is True:
                break
        time.sleep(1)


def until(*callbacks, **kwargs):
    return wait_for(0, 20, *callbacks, **kwargs)


def retry(attempts, *callbacks, **kwargs):
    """
    Repeatedly try callbacks a fixed number of times or until all return True
    """
    for attempt in xrange(attempts):
        if 'bar' in kwargs:
            kwargs['bar'].next(attempt == 0, message=kwargs.get('message'))
        for callback in callbacks:
            if not callback():
                break
        else:
            break
    else:
        raise OSError("Retry attempts exceeded")
    return True


def juju_state_server():
    if api_endpoints() != 0:
        return False
    endpoints = json.loads(sh.juju('api-endpoints', '--format=json'))
    for ep in endpoints:
        host, port = ep.split(':', 1)
        result = socket_open(host, int(port))
        if result is True:
            return ep
    return False


def get_client():
    import jujuclient
    global _client
    if _client is None:
        def env_connection():
            # It seems that calling status forces the rewrite of
            # .jenv with the state server
            sh.juju('status')
            data = yaml.load(open(get_jenv()))
            return data.get('state-servers', []) != []
        current = current_env()
        wait_for(1000, 20, env_connection)
        _client = jujuclient.Environment.connect(current)
    return _client


def cf_service():
    """wait_for help to ensure juju agent state reflects started before
    continuing with other checks that depend on the state of a CF deployment.
    """
    status = get_client().status()
    if 'cloudfoundry' not in status['Services']:
        return False
    sdata = status['Services']['cloudfoundry']
    for unit in sdata['Units'].values():
        if unit['Agent']['Status'] not in ['installed', 'started']:
            return False
    return True


def reconciler_endpoint():
    status = get_client().status()
    try:
        units = status['Services']['cloudfoundry']['Units']
        unit = sorted(units.keys())[0]
        hostname = units[unit]['PublicAddress']
        if not hostname:
            return False
        return hostname
    except (KeyError, IndexError, TypeError):
        return False


def webadmin_endpoint():
    status = get_client().status()
    try:
        units = status['Services']['webadmin']['Units']
        unit = sorted(units.keys())[0]
        hostname = units[unit]['PublicAddress']
        if not hostname:
            return False
        return hostname
    except (KeyError, IndexError, TypeError):
        return False


def endpoint():
    status = get_client().status()
    try:
        units = status['Services']['haproxy']['Units']
        unit = sorted(units.keys())[0]
        hostname = units[unit]['PublicAddress']
        if not hostname:
            return False
        return dig(hostname)
    except (KeyError, IndexError):
        return False
    return endpoint


def api():
    """Return the proper CF api endpoint for connecting to a running
    deployment.
    """
    return "http://api.%s.xip.io" % endpoint()


def bootstrap():
    if not os.path.exists(get_jenv()) or api_endpoints() != 0:
        juju = sh.check('juju', throw=False)
        return juju('bootstrap') != 0
    return True


def get_admin_password():
    jenv = get_jenv()
    data = yaml.load(open(jenv))
    return data['password']


def get_repo_path():
    candidates = [os.getcwd(), os.environ.get('JUJU_REPOSITORY')]
    for repo_path in candidates:
        if not repo_path:
            continue
        if 'trusty' in repo_path:
            repo_path, _ = os.path.split(repo_path)
            if repo_path.endswith('trusty'):
                repo_path = os.path.dirname(repo_path)
        if os.path.exists(os.path.join(repo_path, 'trusty', 'cloudfoundry')):
            return repo_path
    sys.stderr.write(
        'Unable to determine valid repository path.\n'
        'Please set JUJU_REPOSITORY and/or ensure that directory containing\n'
        'the charm includes the series.\n')
    sys.exit(1)


def deploy(**config):
    status = get_client().status()
    constraints = config.pop('constraints', None)
    if 'cloudfoundry' in status['Services']:
        return True
    # create an up to date config
    config = dict({
        'admin_secret': get_admin_password(),
        'cf_version': 'latest',
        'placement': 'dense',
    }, **config)

    fd, fn = tempfile.mkstemp(suffix='.yaml')
    os.close(fd)
    with open(fn, 'w') as fp:
        yaml.dump({'cloudfoundry': config}, fp)

    repo_path = get_repo_path()

    args = ['deploy', '--config=%s' % fn,
            '--repository=%s' % repo_path]
    if constraints:
        args.append('--constraints=%s' % constraints)
    args.append('local:trusty/cloudfoundry')
    juju = sh.check('juju', throw=False)
    if juju(*args) != 0:
        return False
    time.sleep(5)
    if juju('expose', 'cloudfoundry') != 0:
        return False
    os.unlink(fn)
    return True


def login(password):
    """wait_for helper to login and establish a default org/space/target
    within CF.
    """
    try:
        cf = sh.check('cf')
        cf('api', api())
        cf('auth', 'admin', password)
        cf('create-space', '-o', 'juju-org', 'my-space')
        cf('target', '-s', 'my-space', '-o', 'juju-org')
    except subprocess.CalledProcessError:
        return False
    return True


def socket_open(addr='127.0.0.1', port=8888):
    try:
        s = socket.socket()
        s.connect((addr, port))
        s.close()
    except Exception:
        return False
    else:
        return True


def process_stopped(pid):
    try:
        os.kill(pid, 0)
        return False
    except OSError as e:
        if e.errno == 3:
            return True
        else:
            raise


def monit_available(url='http://localhost:2812'):
    try:
        r = requests.get(url)
        return r.ok
    except requests.exceptions.ConnectionError:
        return False


def setup_modprobe(module):
    ''' Load a kernel module and configure for auto-load on reboot '''
    try:
        subprocess.check_call(['modprobe', module])
        with open('/etc/modules', 'r+') as modules:
            if module not in modules.read():
                modules.write(module + '\n')
    except subprocess.CalledProcessError:
        # In certain containers we won't have access to kernel modules
        # continue w/o
        pass


def linux_image_extra_package():
    """
    Get the explicitly tagged LIE package for the current kernel release.

    Occasionally, there is a short window of drift between the
    linux-image-extra-virtual package and the kernel in the cloud image.
    So, instead of relying on the -virtual package, determine the version
    based on the actual kernel release we're running on.
    """
    kernel_release = subprocess.check_output(['/bin/uname', '-r']).strip()
    return 'linux-image-extra-{}'.format(kernel_release)


def install_linux_image_extra():
    pkg = linux_image_extra_package()
    fetch.apt_install(pkg)


def apt_install(package_list):
    return [partial(fetch.apt_install, p) for p in package_list]


def modprobe(mods):
    return [partial(setup_modprobe, mod) for mod in mods]


def serializable(obj):
    if hasattr(obj, 'serialized'):
        if callable(obj.serialized):
            return obj.serialized()
        else:
            return obj.serialized
        return obj


def jsonified(f):
    @wraps
    def wrapper(obj):
        result = f()
        return jsonify(result)
    return wrapper


def jsonify(obj):
    return json.dumps(obj, indent=2, default=serializable, sort_keys=True)


def to_ip(address):
    """Map a string address to a publicly reachable IP address."""
    if not address:
        return None
    if ip_pat.match(address):
        return address  # already an IP
    else:
        return resolve_public_ip(address)


def resolve_public_ip(address):
    """Attempt to resolve the addresses from google public DNS followed by
    machine local DNS."""
    def filter_ip(dig_output):
        for candidate in dig_output.split('\n'):
            candidate = candidate.strip()
            if ip_pat.match(candidate):
                return candidate
        return None

    for server in ['@8.8.8.8', '']:
        try:
            output = subprocess.check_output(
                filter(None, ['dig', '+short', server, address]))
            result = filter_ip(output)
            if result:
                return result
        except subprocess.CalledProcessError:
            pass

    return None


def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None
