#!/usr/bin/env python

import json
import os
import requests
import subprocess
import yaml
from urlparse import urlparse

from charmhelpers.core import hookenv
from charmhelpers.core import services

from charmgen.generator import CharmGenerator
from cloudfoundry.releases import RELEASES
from cloudfoundry.services import SERVICES
from cloudfoundry.contexts import JujuAPICredentials
from cloudfoundry.contexts import OrchestratorRelation
from cloudfoundry.contexts import CloudFoundryCredentials
from cloudfoundry.contexts import RequiredConfig
from cloudfoundry.path import path
from cloudfoundry.utils import wait_for
from cloudfoundry.utils import socket_open

from deployer.utils import setup_logging
from raindance.package import PackageArchive


SOFTWARE = 'cf'
ARCH = 'amd64'  # FIXME: determine from system


@hookenv.cached
def get_version():
    config = hookenv.config()
    version = config.get('cf_version')
    if not version or version == 'latest':
        version = RELEASES[0]['releases'][1]
    return str(version)


def precache_job_artifacts(s):
    config = hookenv.config()
    if not config['mirror_artifacts']:
        return
    version = get_version()
    url = config['artifacts_url']
    assert urlparse(url).scheme != 'nfs', 'NFS URLs not supported with mirror_artifacts'
    pa = PackageArchive(url)
    genmirror = pa.build_mirror_section(path('/srv/data'), SOFTWARE, [(version, ARCH)])
    files = [x for x in genmirror]
    return files


def deploy(s):
    config = hookenv.config()
    version = config.get('cf_version')
    if not version or version == 'latest':
        version = RELEASES[0]['releases'][1]
    from_charmstore = not config['generate_dependents']
    cs_namespace = config['charmstore_namespace']
    admin_ui = config['admin_ui']
    generator = CharmGenerator(RELEASES, SERVICES,
                               from_charmstore, cs_namespace, admin_ui)
    generator.select_release(version)
    placement = generator.load_placement_policy(config['placement'])
    bundle = generator.build_deployment(placement)['cloudfoundry']
    wait_for(300, 5, socket_open)
    hookenv.log(json.dumps(bundle))
    requests.post('http://localhost:8888/api/v1/', data=json.dumps(bundle))


def cache_unit_addresses(s):
    units = {}
    for rid in sorted(hookenv.relation_ids(OrchestratorRelation.name)):
        for unit in sorted(hookenv.related_units(rid)):
            units[unit] = hookenv.relation_get('private-address', unit, rid)
    cache_file = path(hookenv.charm_dir()) / 'unit_addresses'
    cache_file.write_text(yaml.safe_dump(units, default_flow_style=False))


def report_consolidated_health():
    cache_file = path(hookenv.charm_dir()) / 'unit_addresses'
    results = {
        'service': 'cloudfoundry',
        'health': 'pass',
        'state': hookenv.juju_status(),
        'units': {},
    }
    if cache_file.exists():
        units = yaml.safe_load(cache_file.text())
        for unit, address in units.iteritems():
            unit_dash = unit.replace('/', '-')
            output = subprocess.check_output([
                'ssh', '-q', 'root@{}'.format(address),
                '-i', path(hookenv.charm_dir()) / 'orchestrator-key',
                '-o', 'UserKnownHostsFile=/dev/null',
                '-o', 'StrictHostKeyChecking=no',
                ' ; '.join([
                    'export CHARM_DIR=/var/lib/juju/agents/unit-{}/charm'.format(unit_dash),
                    'cd $CHARM_DIR',
                    'hooks/health',
                ]),
            ])
            results['units'][unit] = yaml.safe_load(output)
            health = results['units'][unit]['health']
            if health == 'fail':
                results['health'] = 'fail'
            elif health == 'warn' and results['health'] != 'fail':
                results['health'] = 'warn'
    print yaml.safe_dump(results, default_flow_style=False)


def manage():
    if hookenv.hook_name() == 'health':
        report_consolidated_health()
        return
    manager = services.ServiceManager([
        {
            'service': 'bundle',
            'required_data': [
                JujuAPICredentials(),
                RequiredConfig('artifacts_url') != '',
            ],
            'provided_data': [OrchestratorRelation(),
                              CloudFoundryCredentials()],
            'data_ready': [
                cache_unit_addresses,
                deploy,
            ],
            'start': [],
            'stop': [],
        },
        {
            'service': 'reconciler',
            'ports': [8888],
            'required_data': [
                JujuAPICredentials(),
                RequiredConfig('artifacts_url') != '',
                {'charm_dir': hookenv.charm_dir(),
                 'environment_uuid': os.environ.get('JUJU_ENV_UUID')},
            ],
            'data_ready': [
                services.render_template('reconciler-upstart.conf',
                                         target='/etc/init/reconciler.conf'),
                services.render_template('reconciler-config.conf',
                                         target='/etc/juju-deployer/server.conf'),
            ],
        },
        {
            'service': 'nfs-kernel-server',
            'required_data': [
                RequiredConfig('mirror_artifacts') == True,  # noqa
                RequiredConfig('artifacts_url') != '',
                {'nfs_units': OrchestratorRelation.remote_units()},
            ],
            'data_ready': [
                precache_job_artifacts,
                services.render_template('nfs-exports',
                                         target='/etc/exports'),
            ],
        },
    ])
    manager.manage()


if __name__ == '__main__':
    setup_logging(verbose=True, debug=True)

    manage()
