#!/usr/bin/env python

"""
Run deployer in a loop, then remove any services not
in the expected state.

To experiment with this::

    #activate virtualenv
    . .tox/py27/bin/activate
    python cloudfoundry/reconciler.py \
        --logging=debug --repo=build --port=8888

    # then in another window you can do the follow to play
    # with the REST API
    cd tests
    ./test-server.sh

    This should push a bundle of expected state, HUP the
    server to force a run and then status the system.
    Debug output should show whats happening and


ISSUES:
    local charm version in charm url
        need to probe server still
    reconcile loop:
        should only trigger builds after push/reality change
        execute should happen by queueing the callback
        should listen directly on the websocket and schedule
            rebuilds on change
    no support for unit state currently (auto-retry/replace, etc)
    relation removal still needs work
"""
import json
import yaml
import logging
import os
import signal
import time
import subprocess
import hashlib

import tornado.autoreload
import tornado.httpserver
import tornado.ioloop
import tornado.options
import tornado.process
import tornado.web

from tornado.options import define, options
from cloudfoundry import utils
from cloudfoundry.path import path
from reconciler import strategy
from reconciler.ui.app import MainHandler
from reconciler.ui.app import DashboardIO
from reconciler.ui.app import static_resources

import config

application = None
env_name = None
server = None
db = None
health = {}
last_state_checksum = None


def reconcile():
    # delta state real vs expected
    # build strategy
    # execute strategy inside lock
    if not db.strategy:
        reality = db.real
        db.build_strategy(reality)
    if db.strategy and application.config.runtime.modify:
        db.execute_strategy()


def poll_health():
    for service_name, service in db.real['Services'].iteritems():
        logging.debug("Poll health for %s: %s", service_name, service)
        units = service.get('Units', {}) or {}
        for unit_name, unit in units.iteritems():
            unit_addr = unit.get('PublicAddress')
            unit_state = unit.get('AgentState')
            if unit_addr:
                loop = tornado.ioloop.IOLoop.instance()
                loop.add_callback(check_health, service_name,
                                  unit_name, unit_addr, unit_state)


def check_health(service_name, unit_name, unit_addr, unit_state):
    service = health.setdefault(service_name, {
        'name': service_name,
        'health': 'unknown',
        'units': {}})
    unit = service['units'].setdefault(unit_name, {
        'name': unit_name,
        'health': 'unknown',
        'state': {'message': 'Pending'}})
    try:
        unit_dash = unit_name.replace('/', '-')
        key_file = path(__file__).dirname() / '..' / 'orchestrator-key'
        output = subprocess.check_output([
            'ssh', 'root@{}'.format(unit_addr),
            '-i', key_file,
            '-o', 'UserKnownHostsFile=/dev/null',
            '-o', 'StrictHostKeyChecking=no',
            ' ; '.join([
                'export CHARM_DIR=/var/lib/juju/agents/unit-{}/charm'.format(unit_dash),
                'cd $CHARM_DIR',
                'hooks/health',
            ]),
        ], stderr=subprocess.STDOUT)
        try:
            unit.update(yaml.safe_load(output))
        except (yaml.error.YAMLError, ValueError):
            unit['health'] = 'fail'
            unit['state'] = {'message': 'Unable to parse health: {}'.format(output)}
    except subprocess.CalledProcessError as e:
        unit['state'] = {'message': 'Unable to retrieve health: {}'.format(e.output)}
        if unit_state == 'started':
            unit['health'] = 'pass'
        elif unit_state == 'error':
            unit['health'] = 'fail'
        else:
            unit['health'] = 'warn'

    units_fail = [u['health'] == 'fail' for u in service['units'].values()]
    units_not_pass = [u['health'] != 'pass' for u in service['units'].values()]
    if all(units_fail):
        service['health'] = 'fail'
    elif any(units_not_pass):
        service['health'] = 'warn'
    else:
        service['health'] = 'pass'


def get_current_state():
    """Get db.real plus heatlh info."""
    global last_state_checksum  # sad panda
    state = dict(db.real, health=health)
    checksum = hashlib.md5(json.dumps(state)).hexdigest()
    changed = checksum != last_state_checksum
    last_state_checksum = checksum
    return dict(state, changed=changed)


def poll_current_state():
    state = get_current_state()
    if state['changed']:
        send_current_state()


def send_current_state(message=None):
    DashboardIO.send({'message_type': 'state.current',
                      'message_data': get_current_state()})


def sig_reconcile(sig, frame):
    logging.info("Forcing reconcile loop")
    tornado.ioloop.IOLoop.instance().add_callback(reconcile)


def sig_restart(sig, frame):
    logging.warning('Caught signal: %s', sig)
    tornado.ioloop.IOLoop.instance().add_callback(shutdown)


def shutdown():
    logging.info('Stopping http server')
    server.stop()

    logging.info('Will shutdown in %s seconds ...',
                 config.MAX_WAIT_SECONDS_BEFORE_SHUTDOWN)
    io_loop = tornado.ioloop.IOLoop.instance()

    deadline = time.time() + config.MAX_WAIT_SECONDS_BEFORE_SHUTDOWN

    def stop_loop():
        now = time.time()
        if now < deadline and (io_loop._callbacks or io_loop._timeouts):
            io_loop.add_timeout(now + 1, stop_loop)
        else:
            io_loop.stop()
            logging.info('Shutdown')
    stop_loop()


class StateHandler(tornado.web.RequestHandler):
    def get(self):
        self.add_header('content-type', 'text/json')
        self.write(json.dumps(db.expected, indent=2, default=utils.serializable))

    def post(self):
        db.expected = json.loads(self.request.body)
        tornado.ioloop.IOLoop.instance().add_callback(reconcile)


class StatusHandler(tornado.web.RequestHandler):
    def get(self):
        self.add_header('content-type', 'text/json')
        self.write(json.dumps(db.real, indent=2, default=utils.serializable))


class StrategyHandler(tornado.web.RequestHandler):
    def get(self):
        self.add_header('content-type', 'text/json')
        self.write(utils.jsonify(db.strategy))


class ResetHandler(tornado.web.RequestHandler):
    def post(self):
        db.reset()


class HistoryHandler(tornado.web.RequestHandler):
    def get(self):
        self.add_header('content-type', 'text/json')
        self.write(utils.jsonify(db.history))


class PrevHandler(tornado.web.RequestHandler):
    def get(self):
        self.add_header('content-type', 'text/json')
        self.write(utils.jsonify(db.previous))


class ReconcilerWebApp(tornado.web.Application):
    @property
    def config(self):
        return self.settings['config']


def main():
    define('config', default='/etc/juju-deployer/server.conf', type=path)

    tornado.options.parse_command_line()
    default_config = {
        'server.address': '127.0.0.1',
        'server.port': 8888,
        'credentials.user': 'user-admin',
        'server.repository': 'build',
        'runtime.observe': True,
        'runtime.modify': True,
        'runtime.poll_interval': 20,  # seconds
    }
    # juju should be available on a local deploy but not in production
    if utils.which('juju') and not default_config.get('juju.environment'):
        default_config['juju.environment'] = utils.current_env()
    config = utils.parse_config(options.config, default_config)

    global application
    application = ReconcilerWebApp([
        (r'/static/(.*)', tornado.web.StaticFileHandler,
         {'path': static_resources}),
        (r"/dashboard.ws", DashboardIO),
        (r"/api/v1/", StateHandler),
        (r"/api/v1/status", StatusHandler),
        (r"/api/v1/history", HistoryHandler),
        (r"/api/v1/strategy", StrategyHandler),
        (r"/debug/api/v1/reset", ResetHandler),
        (r"/debug/api/v1/prev", PrevHandler),
        (r"/", MainHandler)
        ],
        autoreload=True,
        config=config,
        **config['server']
    )

    global server
    global db

    if not os.path.exists(config['server.repository']):
        os.makedirs(config['server.repository'])

    db = strategy.StateEngine(config)
    db.recover_state()

    server = tornado.httpserver.HTTPServer(application)
    server.listen(config['server.port'], config['server.address'])

    signal.signal(signal.SIGTERM, sig_restart)
    signal.signal(signal.SIGINT, sig_restart)
    signal.signal(signal.SIGHUP, sig_reconcile)

    DashboardIO.add_listener('state.current', send_current_state)

    # config file change should reload the application
    # its values are passed to Application init
    tornado.autoreload.watch(options.config)
    utils.record_pid()
    loop = tornado.ioloop.IOLoop.instance()
    # call poll_health ASAP, to populate the initial health data, and
    # also schedule poll_health for every 60s to refresh health data
    loop.add_callback(poll_health)
    tornado.ioloop.PeriodicCallback(poll_health, 60000, io_loop=loop).start()
    tornado.ioloop.PeriodicCallback(poll_current_state, 60000,
                                    io_loop=loop).start()
    if config['runtime.poll_interval']:
        tornado.ioloop.PeriodicCallback(reconcile,
                                        config['runtime.poll_interval'] * 100,
                                        io_loop=loop).start()
    loop.start()

if __name__ == "__main__":
    main()
