import hashlib
import json
import logging
import threading

import tornado.ioloop
from tornado import gen

from reconciler import tactics
from reconciler.config import (PENDING, COMPLETE, FAILED, RUNNING)
from cloudfoundry import utils
from cloudfoundry import tsort

from jujuclient import Environment
from deployer.service import Service


class StateEngine(object):
    def __init__(self, config):
        self.config = config
        self._env = None
        # expected is the state we are
        # transitioning to
        self._expected = {}
        # previous is the last (optional)
        # expected state
        self.previous = {}
        self.strategy = Strategy(self.env)
        self.history = []
        self.exec_lock = threading.Lock()

    def reset(self):
        self.expected = {}
        self._reset_strategy()

    def _reset_strategy(self):
        if self.strategy:
            self.history.append(self.strategy)
        self.strategy = Strategy(self.env)

    @property
    def env(self):
        if self._env:
            return self._env
        c = self.config
        api_address = c.get('juju.api_address')
        if api_address:
            api_address = api_address.encode('utf8')
        self._env = self.get_env(
            env_name=c.get('juju.environment'),
            api_address=api_address,
            user=c['credentials.user'],
            password=c['credentials.password'])
        return self._env

    @property
    def expected(self):
        return self._expected

    @expected.setter
    def expected(self, value):
        self._expected = value
        self.env.set_annotation(self.config['juju.environment_uuid'],
                                'environment',
                                {'reconciler-state': json.dumps(self._expected)})

    def recover_state(self):
        """Attempt to read an annotation on the environment of expected state.
        """
        data = self.env.get_annotation(self.config['juju.environment_uuid'],
                                       'environment')
        if data:
            annotations = data.get('Annotations')
            if annotations:
                state = annotations.get('reconciler-state')
                if state:
                    state = json.loads(state)
                    self._expected = state
                    logging.info("Recovering Expected State from Annotation")
                    logging.debug(json.dumps(self._expected, indent=2))

    @property
    def real(self):
        return self.env.status()

    def sort_by_placement(self, service_names):
        dependencies = {}
        if not self.expected:
            return []

        for name in self.expected['services'].keys():
            dependencies[name] = []
            service = Service(name, self.expected['services'][name])
            for placement in service.unit_placement:
                target = placement.split(':')[-1].split('=')[0]
                if not target.isdigit():
                    # only placement onto another named service matter
                    dependencies[name].append(target)
        ordering = tsort.sort(dependencies)
        return sorted(service_names, key=ordering.index)

    def build_strategy(self, reality=None):
        if reality is None:
            reality = self.real
        if reality is None:
            return []

        # Service Deltas
        self.strategy.extend(self.build_services())
        self.strategy.extend(self.build_relations())

    def _changed(self):
        previous = hashlib.md5(json.dumps(self.previous or {}, sort_keys=True))
        expected = hashlib.md5(json.dumps(self.expected, sort_keys=True))
        return expected != previous

    def build_services(self):
        # This should do a 3-way merge from the previous expected state
        # to current with a delta for reality
        current = self.expected
        result = []

        real = self.real
        # Note: This deals with bundle format and the juju-core status output
        # formats, so the key names are a little different
        # XXX This is a very shallow diff in the sense that it only does
        # service name merges and not charm revision
        actual = set(real.get('Services', {}).keys())
        expected = set(current.get('services', {}).keys())
        previous = set((self.previous or {}).get('services', {}).keys())
        adds = expected - actual
        deletes = (previous - expected) & actual

        if self._changed():
            # this is still over-aggressive, but better than always doing it
            result.append(tactics.GenerateTactic(
                repo=self.config['server.repository']))
        for service_name in self.sort_by_placement(adds):
            service = current['services'][service_name].copy()
            service['service_name'] = service_name
            branch = service.get('branch')
            if branch and branch.startswith('local:'):
                result.append(tactics.UpdateCharmTactic(
                    charm_url=branch,
                    repo=self.config['server.repository']))
            result.append(tactics.DeployTactic(service=service,
                                               repo=self.config['server.repository']))
        for service_name in deletes:
            result.append(
                tactics.RemoveServiceTactic(service_name=service_name))

        for service_name in actual:
            # Handle (un)expose
            service = current.get('services', {}).get(service_name, {}).copy()
            service['service_name'] = service_name
            exposed = real.get('Services', {}).get(service_name, {}).get('expose', False)
            expose_intent = service.get('expose', False)
            if expose_intent is True and exposed is False:
                result.append(tactics.ExposeTactic(service=service))
            elif expose_intent is False and exposed is True:
                result.append(tactics.UnexposeTactic(service=service))

        logging.debug("Build Strategy %s", result)
        return result

    def build_relations(self):
        current = self.expected
        result = []
        if not current:
            return result

        real = self.real
        # prev = self.previous

        crels = utils.flatten_relations(current.get('relations', []))
        for rel in crels:
            if not utils.rel_exists(real, rel[0], rel[1]):
                result.append(tactics.AddRelationTactic(
                    endpoint_a=rel[0], endpoint_b=rel[1]))

        # XXX skip deletes for now

        return result

    def execute_strategy(self):
        # each strategy is a list of tactics,
        # we track the state of each of those
        # by mutating the tactic in the list
        # execute strategy shouldn't block
        # so if the lock is taken we continue
        # without modifying the strategy
        # and pick up in the next timeout cycle
        if not self.strategy.runnable:
            return

        if not self.exec_lock.acquire(False):
            return
        try:
            logging.debug("Exec Strat %s", self.strategy)
            self.strategy()
            if self.strategy.state == COMPLETE:
                self.previous = self.expected
            if self.strategy.runnable:
                tornado.ioloop.IOLoop.instance().add_callback(
                    self.execute_strategy)
            else:
                self._reset_strategy()
        finally:
            self.exec_lock.release()

    @classmethod
    def get_env(cls, env_name=None, api_address=None, user=None, password=None):
        if api_address:
            # use the explicit API address option
            env = Environment(api_address)
            env.login(user=user, password=password)
        elif env_name:
            # use the local option which parses local jenv info
            env = Environment.connect(env_name)
        else:
            raise ValueError('Either name or api_address must be given')
        return env


class Strategy(list):
    def __init__(self, env):
        self.state = PENDING
        self.env = env

    def find_next_tactic(self):
        if not self:
            return None
        for tactic in self:
            if tactic.state == FAILED:
                return None
            if tactic.state == PENDING:
                return tactic
        return None

    @property
    def runnable(self):
        return bool(self.find_next_tactic())

    @gen.coroutine
    def __call__(self, env=None):
        if env is None:
            env = self.env

        current = self.find_next_tactic()
        if not current:
            self.state = COMPLETE
        else:
            self.state = RUNNING
            current.run(env)
            if current.state == FAILED:
                self.state = FAILED
            next = self.find_next_tactic()
            if not next and current.state == COMPLETE:
                self.state = COMPLETE

    def __str__(self):
        return "Strategy %s" % [str(t) for t in self]

    def serialized(self):
        return self
