import datetime
import logging
import os
import shutil

from reconciler.config import (
    PENDING, COMPLETE, RUNNING, FAILED, STATES
    )

from deployer.service import Service
from deployer.service import ServiceUnitPlacement
from deployer.utils import get_qualified_charm_url
from deployer.utils import parse_constraints


from charmgen.generator import CharmGenerator
from cloudfoundry.releases import RELEASES
from cloudfoundry.services import SERVICES
from cloudfoundry.path import path


class RetryTacticException(Exception):
    def __init__(self, retry_timeout=None, log=None):
        self.retry_timeout = retry_timeout
        self.log = log


class Tactic(object):
    def __init__(self,  **kwargs):
        self.state = PENDING
        self.failure = None
        self.initial_start_time = None
        self.start_time = None
        self.end_time = None
        self.final_end_time = None
        self.kwargs = kwargs

    def __str__(self):
        return "%s [%s]: %s" % (self.name, STATES[self.state], self.kwargs)

    def serialized(self):
        d = dict(vars(self))
        d.update({"name": self.name})
        return d

    def run(self, env):
        if self.state != PENDING:
            raise ValueError("strategy out of order")
        self.start_time = datetime.datetime.now()
        self.initial_start_time = self.initial_start_time or self.start_time
        self.state = RUNNING
        try:
            logging.debug("Running %s", self)
            self._run(env, **self.kwargs)
            self.state = COMPLETE
        except RetryTacticException as e:
            total_run_time = datetime.datetime.now() - self.initial_start_time
            total_run_time = total_run_time.total_seconds()
            if not e.retry_timeout or total_run_time < e.retry_timeout:
                self.state = PENDING
            else:
                if e.log:
                    logging.log(*e.log.args, **e.log.kwargs)
                logging.error("Tactic retry timeout %s after %d seconds",
                              self, total_run_time)
                self.state = FAILED
                self.failure = 'Retry timeout after %d seconds' % total_run_time
        except Exception as e:
            logging.error("Tactic Failed", exc_info=True)
            self.state = FAILED
            self.failure = str(e)
        finally:
            self.end_time = datetime.datetime.now()
            if self.state != PENDING:
                self.final_end_time = self.end_time


class GenerateTactic(Tactic):
    name = "Generate charms"

    def _run(self, env,  **kwargs):
        version = kwargs.get('cf_release',  RELEASES[0]['releases'][1])
        build_dir = os.path.join(kwargs['repo'], str(version))
        if os.path.exists(build_dir):
            return
        generator = CharmGenerator(RELEASES, SERVICES, from_charmstore=False)
        generator.select_release(version)
        generator.generate(build_dir)


class UpdateCharmTactic(Tactic):
    name = "Update charm"

    def _run(self, env, **kwargs):
        charm_url = get_qualified_charm_url(kwargs['charm_url'])
        if charm_url.startswith('local:'):
            series, charm_id = charm_url.split(':')[1].split('/')
            charm_name = charm_id.rsplit('-', 1)[0]
            archive = path('/tmp') / charm_name
            version = kwargs.get('cf_release',  RELEASES[0]['releases'][1])
            charm_path = path(kwargs['repo']) / str(version) / series / charm_name
            shutil.make_archive(archive, 'zip', charm_path)
            archive = archive + '.zip'
            size = os.path.getsize(archive)
            with open(archive) as fp:
                env.add_local_charm(fp, series, size)
        else:
            env.add_charm(charm_url)


class DeployTactic(Tactic):
    name = "Deploy"

    class FakeDeployment(object):
        class LogCapture(object):
            def __init__(self):
                self.args = ()
                self.kwargs = {}

            def warning(self, *args, **kwargs):
                self.args = (logging.WARNING,) + args
                self.kwargs = kwargs

            def error(self, *args, **kwargs):
                self.args = (logging.ERROR,) + args
                self.kwargs = kwargs

        log = LogCapture()

    def _run(self, env, **kwargs):
        s = kwargs['service']
        svc = Service(s['service_name'], s)
        charm_url = s.get('branch', s.get('charm'))
        is_local = charm_url.startswith('local:')
        if is_local:
            # XXX version hack, have to probe for actual version after push
            charm_url = charm_url + "-0"
        else:
            charm_url = get_qualified_charm_url(charm_url)
        constraints = None
        if svc.constraints:
            constraints = parse_constraints(svc.constraints)
        status = env.get_stat()
        dep = self.FakeDeployment()
        placement = ServiceUnitPlacement(svc, dep, status)
        targets = [t.split(':')[-1].split('=')[0] for t in svc.unit_placement]
        missing_units = any('units' not in status['services'].get(s, {}) for s in targets)
        if missing_units:
            raise RetryTacticException(retry_timeout=60, log=dep.log)
        machine_spec = placement.get(0)
        if targets and machine_spec is None:
            raise RetryTacticException(retry_timeout=60, log=dep.log)
        env.deploy(svc.name,
                   charm_url,
                   config=svc.config,
                   constraints=constraints,
                   num_units=svc.num_units,
                   machine_spec=machine_spec)


class ExposeTactic(Tactic):
    name = "Expose Service"

    def _run(self, env, **kwargs):
        s = kwargs['service']
        env.expose(s['service_name'])


class UnexposeTactic(Tactic):
    name = "Unexpose Service"

    def _run(self, env, **kwargs):
        s = kwargs['service']
        env.unexpose(s['service_name'])


class RemoveServiceTactic(Tactic):
    name = "Remove Service"

    def _run(self, env, **kwargs):
        env.destroy_service(kwargs['service_name'])


class AddRelationTactic(Tactic):
    name = "Add Relation"

    def _run(self, env, **kwargs):
        env.add_relation(kwargs['endpoint_a'], kwargs['endpoint_b'])


class RemoveRelationTactic(Tactic):
    name = "Add Relation"

    def _run(self, env, **kwargs):
        env.remove_relation(kwargs['endpoint_a'], kwargs['endpoint_b'])
