#!/usr/bin/env python
import argparse
import json
import logging
import os
import subprocess
import sys
import tempfile

from cloudfoundry.releases import RELEASES
from cloudfoundry.utils import (api,
                                bootstrap,
                                cd,
                                cf_service,
                                deploy,
                                endpoint,
                                login,
                                sh,
                                wait_for)
import requests

options = None


def setup():
    global options
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--version', default="latest")
    options = parser.parse_args()
    if options.version == 'latest':
        options.version = RELEASES[0]['releases'][1]
    return options


def validate_app(endpoint, url='/', status_code=200, expected=None):
    url = endpoint + url
    r = requests.get(url)
    assert r.status_code == status_code, "Invalid HTTP status code"
    if expected:
        assert r.text == expected, "Invalid result, got %s, expected %s" % (
            r.text, expected
        )


def github(path, owner="cloudfoundry", repo="cf-release"):
    return requests.get(os.path.join(
        'https://api.github.com/repos',
        owner, repo, 'git', path)).json()


def get_cats(version):
    logging.info("Getting CATs from github")
    if not os.path.exists('cf-acceptance-tests'):
        sh.git('clone',
               'https://github.com/cloudfoundry/cf-acceptance-tests.git')
    with cd('cf-acceptance-tests'):
        sh.git('fetch')
        sh.check('./bin/compile')
        # Switch to the branch of the revision in question
        # Use github api to pull rev/tag under test
        # Use that to get the SHA of that versions CATs
        # switch the branch to that commit
        tag = github('refs/tags/v%s' % version)
        tree = github('trees/%s?recursive=1' % tag['object']['sha'])
        acceptance = [d for d in tree['tree'] if d['path'].startswith('src/acceptance')]
        sha = acceptance[0]['sha']
        print "Switching CATs to {} to test version {}".format(
            sha[:6], version)
        sh.git('reset', '--hard', sha)


def run_cats():
    ep = endpoint()
    with cd('cf-acceptance-tests'):
        fd, fn = tempfile.mkstemp(suffix=".json")
        os.close(fd)
        with open(fn, 'w') as fp:
            defaults = {
                "suite_name": "smoke",
                "apps_domain": "%s.xip.io" % ep,
                "api": api(),
                "syslog_drain_port": 8082,
                "syslog_ip_address": ep,
                "password": "admin",
                "admin_password": "admin",
                "user": "admin",
                "admin_user": "admin",
                "org": "juju-org",
                "space": "my-space",
                "use_existing_org": True,
                "use_existing_space": True,
                "skip_ssl_validation": True,
            }
            json.dump(defaults, fp, indent=2, sort_keys=True)
        logging.info("Running CATs %s", defaults)
        env = os.environ.copy()
        env.update({'CONFIG': os.path.abspath(fn)})
        subprocess.call('./bin/test', shell=True, env=env)


def main():
    options = setup()
    logging.basicConfig(level=logging.INFO)
    get_cats(options.version)
    bootstrap()
    deploy()
    wait_for(60 * 40, 30, cf_service, endpoint, login)
    run_cats()
    sys.exit(0)


if __name__ == '__main__':
    main()
