#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import shutil
import tempfile
import unittest
from contextlib import contextmanager

import yaml
import mock

from charmgen.generator import CharmGenerator, main
from cloudfoundry.contexts import OrchestratorRelation
from cloudfoundry.path import path

from cloudfoundry import contexts

# Local fixture
from tests.release1 import RELEASES, SERVICES


@contextmanager
def tempdir():
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d)


class TestGenerator(unittest.TestCase):

    def test_select_release(self):
        g = CharmGenerator(RELEASES, SERVICES)
        self.assertRaises(KeyError, g.select_release, 'anything')
        self.assertRaises(KeyError, g.select_release, 153)

        self.assertEquals(g.select_release(171), RELEASES[-1])
        self.assertEquals(g.select_release('171'), RELEASES[-1])
        self.assertEquals(g.select_release(172), RELEASES[-1])
        self.assertEquals(g.select_release(173), RELEASES[0])
        # Tests the open range
        self.assertEquals(g.select_release(199), RELEASES[0])
        self.assertEquals(g.release, RELEASES[0])
        self.assertEquals(g.release_version, 199)

    def test_build_metadata(self):
        g = CharmGenerator(RELEASES, SERVICES)
        g.select_release(173)
        cc = g.release['topology']['services'][0][0]
        meta = g.build_metadata(cc)
        self.assertEqual(meta['name'], 'cloud_controller_v1')
        self.assertEqual(meta['author'], CharmGenerator.author)
        cc_service = SERVICES[cc]
        self.assertEqual(meta['summary'], cc_service['summary'])
        self.assertEqual(meta['description'], cc_service['description'])

        # Interface generation
        provides = meta['provides']
        requires = meta['requires']
        self.assertEqual(provides[contexts.CloudControllerRelation.name][
            'interface'], contexts.CloudControllerRelation.interface)
        self.assertEqual(requires[contexts.NatsRelation.name]['interface'],
                         contexts.NatsRelation.interface)
        self.assertEqual(requires[contexts.OrchestratorRelation.name]['interface'],
                         contexts.OrchestratorRelation.interface)
        # This interface is added implicitly by our model, generated
        # charms have a relation to their Orchestrator.
        self.assertEqual(requires[OrchestratorRelation.name][
            'interface'], OrchestratorRelation.interface)

    def test_build_hooks(self):
        g = CharmGenerator(RELEASES, SERVICES)
        g.select_release(173)
        cc = g.release['topology']['services'][0]
        hooks = g.build_hooks(cc)
        self.assertIn('db-relation-changed', hooks)
        self.assertIn('nats-relation-joined', hooks)
        self.assertIn('orchestrator-relation-joined', hooks)
        self.assertIn('install', hooks)
        self.assertIn('start', hooks)

    def test_build_entry_point(self):
        g = CharmGenerator(RELEASES, SERVICES)
        g.select_release(173)
        cc = g.release['topology']['services'][0]
        entry = g.build_entry(cc)
        self.assertIn('job_manager("cloud_controller_v1")', entry)

    def test_generate_charm(self):
        g = CharmGenerator(RELEASES, SERVICES)
        g.select_release(173)
        cc = g.release['topology']['services'][0]
        with tempdir() as tmpdir:
            g.generate_charm(cc, os.path.join(tmpdir, 'build'))
            self.assertTrue(os.path.exists(
                os.path.join(tmpdir, 'build', 'metadata.yaml')))
            self.assertTrue(os.path.isdir(
                os.path.join(tmpdir, 'build', 'hooks')))
            self.assertTrue(os.path.islink(
                os.path.join(tmpdir, 'build', 'hooks', 'db-relation-changed')))
            self.assertTrue(os.path.isfile(
                os.path.join(tmpdir, 'build', 'hooks', 'entry.py')))
            self.assertTrue(os.path.isfile(
                os.path.join(tmpdir, 'build', 'hooks', 'entry.py')))
            self.assertTrue(os.access(
                os.path.join(tmpdir, 'build', 'hooks', 'entry.py'),
                os.R_OK | os.X_OK))

    def test_get_relations(self):
        releases = [{
            'releases': (1,),
            'topology': {
                'services': [('service1', 's1'), ('service2', 's2'), 'cs:trusty/service3'],
                'relations': [('s2:etcd', 'etcd:client')],
            },
        }]
        services = {
            'service1': {'jobs': [{
                'required_data': [contexts.NatsRelation, contexts.MysqlRelation],
            }]},
            'service2': {'jobs': [{
                'provided_data': [contexts.NatsRelation],
                'required_data': [contexts.LTCRelation, contexts.EtcdRelation],
            }]},
        }
        g = CharmGenerator(releases, services)
        g.select_release(1)
        self.assertItemsEqual(g._get_relations(), [
            ('s2:etcd', 'etcd:client'),
            (('s1', 'nats'), ('s2', 'nats')),
            (('s1', 'orchestrator'), ('cloudfoundry', 'orchestrator')),
            (('s2', 'orchestrator'), ('cloudfoundry', 'orchestrator')),
        ])

    def test_build_charm_ref_local(self):
        g = CharmGenerator(RELEASES, SERVICES, from_charmstore=False)
        self.assertEqual(g._build_charm_ref('cs:trusty/mysql'),
                         {'charm': 'cs:trusty/mysql'})
        self.assertEqual(g._build_charm_ref('cloud_controller_v1'),
                         {'charm': 'cloud_controller_v1',
                          'branch': 'local:trusty/cloud_controller_v1'})

    def test_build_charm_ref_cs(self):
        g = CharmGenerator(RELEASES, SERVICES, from_charmstore=True)
        self.assertEqual(g._build_charm_ref('cs:trusty/mysql'),
                         {'charm': 'cs:trusty/mysql'})
        self.assertEqual(g._build_charm_ref('cloud_controller_v1'),
                         {'charm': 'cs:~cf-charmers/trusty/cloud_controller_v1'})

    def test_build_charm_ref_cs_recommended(self):
        g = CharmGenerator(RELEASES, SERVICES, from_charmstore=True, cs_namespace='')
        self.assertEqual(g._build_charm_ref('cs:trusty/mysql'),
                         {'charm': 'cs:trusty/mysql'})
        self.assertEqual(g._build_charm_ref('cloud_controller_v1'),
                         {'charm': 'cs:trusty/cloud_controller_v1'})

    def test_parse_charm_ref(self):
        g = CharmGenerator(RELEASES, SERVICES)
        self.assertEqual(g._parse_charm_ref('cs:trusty/mysql'),
                         ('cs:trusty/mysql', 'mysql', 'mysql'))
        self.assertEqual(g._parse_charm_ref(('cs:trusty/mysql', 'db')),
                         ('cs:trusty/mysql', 'mysql', 'db'))
        self.assertEqual(g._parse_charm_ref(('cloud_controller_v1', 'cc')),
                         ('cloud_controller_v1', 'cloud_controller_v1', 'cc'))

    def test_normalize_relations(self):
        g = CharmGenerator(RELEASES, SERVICES)
        self.assertEqual(g._normalize_relation(('cc', 'db')), 'cc:db')
        self.assertEqual(g._normalize_relation('cc:db'), 'cc:db')

    def test_load_placement_policy(self):
        g = CharmGenerator(RELEASES, SERVICES)
        self.assertEqual(g.load_placement_policy('sparse'), {})
        self.assertEqual(g.load_placement_policy('local'), {
            '__default__': {'constraints': 'arch=amd64 root-disk=20G mem=12G'},
            'dea': {'to': 'kvm:0'},
        })
        self.assertEqual(g.load_placement_policy('manual:zones: {a: [cc]}'), {
            'zones': {'a': ['cc']},
        })
        self.assertRaises(KeyError, g.load_placement_policy, 'unknown')

    def test_process_zones(self):
        g = CharmGenerator(RELEASES, SERVICES)
        services = dict(zip('ABCDE', [{} for n in range(5)]))
        zones = {
            'a': ['A', 'D'],
            'b': ['B', 'C'],
        }
        g._process_zones(services, zones)
        self.assertEqual(services, {
            'A': {},
            'B': {},
            'C': {'to': ['B']},
            'D': {'to': ['A']},
            'E': {},
        })

    def test_process_zones_trailing_default(self):
        g = CharmGenerator(RELEASES, SERVICES)
        services = dict(zip('ABCDE', [{} for n in range(5)]))
        zones = {
            'a': ['A', 'D'],
            'b': ['B', '*'],
        }
        g._process_zones(services, zones, {'constraints': 'mem=10G'})
        self.assertEqual(services, {
            'A': {},
            'B': {'constraints': 'mem=10G'},
            'C': {'to': ['B']},
            'D': {'to': ['A']},
            'E': {'to': ['B']},
        })

    def test_process_zones_only_default(self):
        g = CharmGenerator(RELEASES, SERVICES)
        _services = dict(zip('ABCDE', [{} for n in range(5)]))
        services = mock.MagicMock(spec_set=dict)  # to force ordering in iteritems()
        services.__getitem__.side_effect = _services.__getitem__
        services.iteritems.return_value = [(k, _services[k]) for k in sorted(_services.keys())]
        zones = {
            'a': ['A', 'D'],
            'b': ['*'],
        }
        g._process_zones(services, zones)
        self.assertEqual(_services, {
            'A': {},
            'B': {},
            'C': {'to': ['B']},
            'D': {'to': ['A']},
            'E': {'to': ['B']},
        })

    def test_process_zones_only_default_single(self):
        g = CharmGenerator(RELEASES, SERVICES)
        _services = dict(zip('ABCDE', [{} for n in range(5)]))
        services = mock.MagicMock(spec_set=dict)  # to force ordering in iteritems()
        services.iteritems.return_value = [(k, _services[k]) for k in sorted(_services.keys())]
        zones = {
            'a': ['*'],
        }
        g._process_zones(services, zones)
        self.assertEqual(_services, {
            'A': {},
            'B': {'to': ['A']},
            'C': {'to': ['A']},
            'D': {'to': ['A']},
            'E': {'to': ['A']},
        })

    def test_do_placement_zones(self):
        g = CharmGenerator(RELEASES, SERVICES)
        g._process_zones = mock.Mock()
        services = dict(zip('ABCDE', [{} for n in range(5)]))
        g._do_placement(services, {'zones': True})
        g._process_zones.assert_called_once_with(services, True, {})
        self.assertEqual(services, {
            'A': {},
            'B': {},
            'C': {},
            'D': {},
            'E': {},
        })

    def test_do_placement_services(self):
        g = CharmGenerator(RELEASES, SERVICES)
        g._process_zones = mock.Mock()
        services = dict(zip('ABCDE', [{} for n in range(5)]))
        g._do_placement(services, {'B': {'to': ['A']}})
        assert not g._process_zones.called
        self.assertEqual(services, {
            'A': {},
            'B': {'to': ['A']},
            'C': {},
            'D': {},
            'E': {},
        })

    def test_build_deployment(self):
        g = CharmGenerator(RELEASES, SERVICES, from_charmstore=False)
        g.select_release(173)
        data = g.build_deployment(None)
        cf = data['cloudfoundry']
        self.assertEqual(cf['series'], 'trusty')
        # A local generated charm
        self.assertEqual(cf['services']['cc'],
                         {'charm': 'cloud_controller_v1',
                          'branch': 'local:trusty/cloud_controller_v1',
                          'constraints': 'root-disk=10G'})
        # A store charm
        self.assertEqual(cf['services']['mysql'],
                         {'charm': 'cs:trusty/mysql',
                          'constraints': 'arch=amd64'})

        # Relations
        rels = set(map(tuple, tuple(cf['relations'])))
        expected = ('mysql:db', ('cc:db',))
        # Verify we found expected relations
        self.assertIn(expected, rels)

    def test_generate_deployment(self):
        g = CharmGenerator(RELEASES, SERVICES)
        g.select_release(173)
        with tempdir() as tmpdir:
            g.generate_deployment(os.path.join(tmpdir, 'build'), None)
            bundle = os.path.join(tmpdir, 'build', 'bundles.yaml')
            self.assertTrue(os.path.exists(bundle))
            with open(bundle) as fp:
                data = yaml.safe_load(fp)
                cf = data['cloudfoundry']
                self.assertEqual(cf['series'], 'trusty')

    def test_generate(self):
        g = CharmGenerator(RELEASES, SERVICES, from_charmstore=False)
        g.select_release(173)
        with tempdir() as tmpdir:
            g.generate(tmpdir, None)
            self.assertTrue(os.path.exists(
                os.path.join(tmpdir, 'bundles.yaml')))
            self.assertTrue(os.path.exists(os.path.join(
                tmpdir, 'trusty', 'cloud_controller_v1')))
            # While generate doesn't create a mysql charm (it's from cs)
            # deployer will eventually place it in this dir pre-deploy
            self.assertFalse(os.path.exists(os.path.join(
                tmpdir, 'trusty', 'mysql')))
            self.assertTrue(os.path.isdir(os.path.join(
                tmpdir, 'trusty', 'cloud_controller_v1',
                'hooks', 'cloudfoundry')))
            self.assertTrue(os.path.isdir(os.path.join(
                tmpdir, 'trusty', 'cloud_controller_v1',
                'hooks', 'charmhelpers')))
            self.assertTrue(os.path.isdir(os.path.join(
                tmpdir, 'trusty', 'cloud_controller_v1', 'files')))

    def test_generate_missing_service(self):
        releases = [{'releases': (1, 1), 'topology': {
            'services': [('missing', '??')],
            'relations': [],
        }}]
        services = {}
        g = CharmGenerator(releases, services)
        g.select_release(1)
        with tempdir() as tmpdir:
            self.assertRaises(KeyError, g.generate, tmpdir, None)

    def test_main(self):
        with tempdir() as tmpdir:
            main(['-d', tmpdir, '173'])
            self.assertTrue((path(tmpdir) / 'bundles.yaml').exists())
            self.assertTrue((path(tmpdir) / 'trusty/nats-v1').exists())

if __name__ == '__main__':
    unittest.main()
