#!/usr/bin/env python

import sys
import time
import json
import unittest
import requests
import subprocess
from pkg_resources import resource_filename
from cloudfoundry.path import path


class TestReconcilerServer(unittest.TestCase):
    server_start_timeout = 15
    server = resource_filename('cloudfoundry', 'reconciler.py')
    here = path(__file__).parent
    maxDiff = None

    @classmethod
    def url(cls, page=''):
        return "http://localhost:8888/api/v1/{}".format(page)

    @classmethod
    def setUpClass(cls):
        """
        Start the reconciler server in a background process, and wait for it
        to start up and begin handling requests.

        If a Juju environment is not bootstrapped, skip all tests.
        """
        try:
            subprocess.check_output(['juju', 'status'], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            if 'environment is not bootstrapped' in e.output:
                raise unittest.SkipTest('Skipping due to no bootstrapped Juju environment')
            raise

        cls.server_proc = subprocess.Popen(
            [sys.executable, cls.server, '--config={}'.format(cls.here / 'server.conf')])
        start = time.time()
        while time.time() - start < cls.server_start_timeout:
            try:
                requests.head(cls.url())
                break
            except requests.ConnectionError:
                time.sleep(1)
        else:
            cls.server_proc.kill()
            cls.server_proc.wait()
            raise requests.ConnectionError('Timed out waiting for server to start')

    @classmethod
    def tearDownClass(cls):
        """
        Gracefully shutdown the background reconciler server.
        """
        cls.server_proc.kill()
        cls.server_proc.wait()

    def setUp(self):
        resp = requests.get(self.url('reset'))
        resp.raise_for_status()

    def api(self, page='', data=None):
        if data is None:
            resp = requests.get(self.url(page))
        else:
            resp = requests.post(self.url(page), data=json.dumps(data))
        resp.raise_for_status()
        if resp.content:
            return resp.json()
        return None

    def normalize_strategy(self, strategy):
        result = []
        for action in strategy:
            charm = action['kwargs'].get('charm_url')
            if 'service' in action['kwargs']:
                service = action['kwargs']['service']
                charm = service.get('charm', service.get('branch'))
            if charm:
                result.append((action['name'], charm))
            else:
                result.append(action['name'])
        return result

    def test_status_empty(self):
        self.assertEqual(self.api(), {})

    def test_set_state(self):
        state = json.loads((self.here / 'state.json').text())
        self.api(data=state)
        self.assertEqual(self.api(), state)

    def test_reset(self):
        state = json.loads((self.here / 'state.json').text())
        self.api(data=state)
        self.api('reset')
        self.assertEqual(self.api(), {})

    def test_strategy(self):
        # can't test transitions because server is configured not to act
        state = {'services': {'s1': {'charm': 'cs:s1'}, 's2': {'branch': 'local:s2'}}}
        self.assertEqual(self.api('strategy'), [])
        self.api(data=state)
        strat = self.normalize_strategy(self.api('strategy'))
        self.assertIn('Generate charms', strat)
        self.assertIn(('Deploy', 'cs:s1'), strat)
        self.assertIn(('Update charm', 'local:s2'), strat)
        self.assertIn(('Deploy', 'local:s2'), strat)


if __name__ == '__main__':
    unittest.main()
