import logging
import mock
import os
import shutil
import tempfile
import unittest

from reconciler import tactics

# Set the log level high here to avoid messages from deployer in the test
# output. They are an artifact of partial specification for testing.
logging.basicConfig(level=logging.CRITICAL)


class TestGenerate(unittest.TestCase):
    def test_generate(self):
        d = tempfile.mkdtemp()
        g = tactics.GenerateTactic()
        g._run(None, cf_release=173, repo=d)
        self.assertTrue(
            os.path.exists(os.path.join(d, '173')))
        self.assertTrue(
            os.path.exists(os.path.join(d, '173', 'trusty')))
        shutil.rmtree(d)


class TestUpgradeCharm(unittest.TestCase):
    @mock.patch.object(tactics, 'get_qualified_charm_url')
    @mock.patch('os.path.getsize')
    @mock.patch('shutil.make_archive')
    @mock.patch('jujuclient.Environment')
    def test_upgrade_local(self, menv, marchive, msize, murl):
        t = tactics.UpdateCharmTactic()
        murl.side_effect = lambda u: u
        mopen = mock.mock_open()
        with mock.patch.object(tactics, 'open', mopen, create=True):
            t._run(menv, charm_url='local:trusty/dea-v1',
                   repo='build',
                   cf_release=173)
            assert mopen.called
        menv.add_local_charm.assert_called_once()
        marchive.assert_called_once()

    @mock.patch.object(tactics, 'get_qualified_charm_url')
    @mock.patch('os.path.getsize')
    @mock.patch('shutil.make_archive')
    @mock.patch('jujuclient.Environment')
    def test_upgrade_store(self, menv, marchive, msize, murl):
        t = tactics.UpdateCharmTactic()
        murl.side_effect = lambda u: u
        mopen = mock.mock_open()
        with mock.patch.object(tactics, 'open', mopen, create=True):
            t._run(menv, charm_url='cs:trusty/dea-v1',
                   repo='build',
                   cf_release=173)
            assert not mopen.called
            assert not menv.add_local_charm.called
            assert menv.add_charm.called
            assert not marchive.called


class TestDeploy(unittest.TestCase):
    @mock.patch('jujuclient.Environment')
    def test_deploy(self, menv):
        t = tactics.DeployTactic()
        t._run(
            menv, service=dict(
                service_name='dea',
                branch='local:trusty/dea-v1',
            ), repo='build')
        menv.deploy.assert_called_once_with(
            'dea', 'local:trusty/dea-v1-0',
            config=None, num_units=1, constraints=None,
            machine_spec=None)

    @mock.patch.object(tactics, 'get_qualified_charm_url')
    @mock.patch('jujuclient.Environment')
    def test_deploy_complex(self, menv, murl):
        menv.get_stat.return_value = {
            'services': {'uaa': {'units': {
                'uaa/1': {'machine': 11},
                'uaa/0': {'machine': 10},
                'uaa/2': {'machine': 12},
            }}}
        }
        murl.return_value = 'url'

        t = tactics.DeployTactic()
        t._run(
            menv, service=dict(
                service_name='dea',
                charm='cs:trusty/dea-v1',
                options={'foo': 'bar'},
                constraints='root-disk=12G',
                to=['kvm:uaa'],
            ), repo='build')
        murl.assert_called_once_with('cs:trusty/dea-v1')
        menv.deploy.assert_called_once_with(
            'dea', 'url',
            config={'foo': 'bar'}, num_units=1,
            constraints={'root-disk': 12288},
            machine_spec='kvm:10')

    @mock.patch.object(tactics, 'get_qualified_charm_url')
    @mock.patch('jujuclient.Environment')
    def test_deploy_retry(self, menv, murl):
        uaa = {}
        menv.get_stat.return_value = {'services': {'uaa': uaa}}
        murl.return_value = 'url'
        service = dict(
            service_name='dea',
            charm='cs:trusty/dea-v1',
            options={'foo': 'bar'},
            constraints='root-disk=12G',
            to=['kvm:uaa'],
        )

        t = tactics.DeployTactic()
        self.assertRaises(tactics.RetryTacticException,
                          t._run, menv, service=service, repo='build')
        uaa['units'] = {}
        self.assertRaises(tactics.RetryTacticException,
                          t._run, menv, service=service, repo='build')
        uaa['units'] = {'uaa/0': {'machine': 10}}
        t._run(menv, service=service, repo='build')
        murl.assert_called_with('cs:trusty/dea-v1')
        menv.deploy.assert_called_once_with(
            'dea', 'url',
            config={'foo': 'bar'}, num_units=1,
            constraints={'root-disk': 12288},
            machine_spec='kvm:10')


class TestRemoveService(unittest.TestCase):
    @mock.patch('jujuclient.Environment')
    def test_remove_service(self, menv):
        t = tactics.RemoveServiceTactic()
        t._run(menv, service_name="foo")
        menv.destroy_service.assert_called_once_with('foo')


class TestAddRelation(unittest.TestCase):
    @mock.patch('jujuclient.Environment')
    def test_add_relation(self, menv):
        t = tactics.AddRelationTactic()
        t._run(menv, endpoint_a="foo", endpoint_b="bar")
        menv.add_relation.assert_called_once_with('foo', 'bar')


class TestRemoveRelation(unittest.TestCase):
    @mock.patch('jujuclient.Environment')
    def test_remove_relation(self, menv):
        t = tactics.RemoveRelationTactic()
        t._run(menv, endpoint_a="foo", endpoint_b="bar")
        menv.remove_relation.assert_called_once_with('foo', 'bar')
