import unittest
import mock

from charmhelpers.core import services
from cloudfoundry.path import path
from cloudfoundry import tasks


class TestTasks(unittest.TestCase):
    def setUp(self):
        self.charm_dir_patch = mock.patch(
            'charmhelpers.core.hookenv.charm_dir')
        self.charm_dir = self.charm_dir_patch.start()
        self.charm_dir.return_value = 'charm_dir'
        self.log_patch = mock.patch(
            'charmhelpers.core.hookenv.log')
        self.log = self.log_patch.start()

    def tearDown(self):
        self.charm_dir_patch.stop()

    @mock.patch('cloudfoundry.utils.wait_for', mock.Mock())
    @mock.patch('charmhelpers.core.host.adduser')
    @mock.patch('subprocess.check_output')
    @mock.patch('subprocess.check_call')
    @mock.patch('charmhelpers.fetch.filter_installed_packages')
    @mock.patch('charmhelpers.fetch.apt_install')
    def test_install_base_dependencies(self, apt_install,
                                       filter_installed_packages,
                                       check_call, check_output, adduser):
        filter_installed_packages.side_effect = lambda a: a
        with mock.patch('cloudfoundry.tasks.path', spec=path) as monitrc:
            monitrc().exists.return_value = False
            tasks.install_base_dependencies()

        apt_install.assert_called_once()
        adduser.assert_called_once_with('vcap')
        assert monitrc.called
        assert monitrc.call_args == mock.call('/etc/monit/conf.d/enable_http')
        newtext = '\nset httpd port 2812 and\n'\
            '   use address localhost\n'\
            '   allow localhost\n'
        assert monitrc().write_text.call_args == mock.call(newtext)
        check_call.assert_has_calls([
            mock.call(['service', 'monit', 'force-reload'], -2),
            mock.call(['gem', 'install',
                       '--no-ri', '--no-rdoc',
                       'charm_dir/files/' +
                       'bosh-template-1.2611.0.pre.gem'])])

    def test_monit_http_enable_idem(self):
        with mock.patch('cloudfoundry.tasks.path', spec=path) as confd:
            confd().exists.return_value = True
            tasks.enable_monit_http_interface()
            assert not confd().write_text.called

    @mock.patch('raindance.package.PackageArchive')
    @mock.patch('cloudfoundry.contexts.OrchestratorRelation')
    def test_fetch_job_artifacts(self, OrchRelation, PackageArchive):
        OrchRelation.return_value = {'orchestrator': [{
            'cf_version': 'version',
            'artifacts_url': 'http://url',
        }]}
        tasks.fetch_job_artifacts('job_name')
        PackageArchive.assert_called_with('http://url')
        PackageArchive().build_mirror_section.assert_called()
        PackageArchive().build_mirror_section.iter.assert_called()

    @mock.patch('raindance.package.PackageArchive')
    @mock.patch('cloudfoundry.contexts.OrchestratorRelation')
    def test_fetch_job_artifacts_nfs(self, orch, PA):
        orch.return_value = {'orchestrator': [{
            'cf_version': 'version',
            'artifacts_url': 'nfs://url',
        }]}
        tasks.fetch_job_artifacts('job_name')
        assert not PA.called

    @mock.patch('raindance.package.PackageArchive')
    @mock.patch('cloudfoundry.contexts.OrchestratorRelation')
    def test_install_job_artifacts(self, OrchRelation, PackageArchive):
        OrchRelation.return_value = {'orchestrator': [{
            'cf_version': 'version',
            'artifacts_url': 'url',
        }]}
        bd = mock.MagicMock(spec=path)
        rd = mock.MagicMock(spec=path)
        pa = PackageArchive.return_value
        ls = rd.__div__().__div__.return_value
        lf = bd.__div__.return_value
        pa.setup_job.return_value = ['job_file']

        tasks.install_job_artifacts(bd, rd, 'job_name')
        PackageArchive.assert_called_with('url')
        pa.setup_job.assert_called()
        lf.unlink_p.assert_called()
        lf.makedirs_p.assert_called()
        ls.symlink.assert_called_with(lf.unlink_p())

    @mock.patch('cloudfoundry.tasks.open', create=True)
    @mock.patch('cloudfoundry.tasks.yaml.safe_load')
    def test_load_spec(self, safe_load, mopen):
        safe_load.side_effect = [
            {'p1': 'v1'},
            {'p2': 'v2'},
        ]
        self.assertEquals(tasks.load_spec('job1'), {'p1': 'v1'})
        self.assertEquals(tasks.load_spec('job1'), {'p1': 'v1'})
        self.assertEqual(mopen.call_count, 1)
        self.assertEqual(safe_load.call_count, 1)
        self.assertEquals(tasks.load_spec('job2'), {'p2': 'v2'})
        self.assertEqual(mopen.call_count, 2)
        self.assertEqual(safe_load.call_count, 2)
        self.assertEqual(mopen.call_args_list, [
            mock.call('/var/vcap/jobs/job1/spec'),
            mock.call('/var/vcap/jobs/job2/spec'),
        ])

    @mock.patch('cloudfoundry.tasks.load_spec')
    @mock.patch('cloudfoundry.templating.RubyTemplateCallback')
    def test_job_templates(self, RubyTemplateCallback, load_spec):
        spec = load_spec.return_value = {'templates': {
            'src1': 'dest1',
            'src2': 'dest2',
        }}
        manager = mock.Mock()
        generated_callbacks = RubyTemplateCallback.side_effect = [
            mock.Mock(), mock.Mock(services.ManagerCallback()), mock.Mock()]
        tasks.job_templates('map')(manager, 'job_name', 'event_name')
        generated_callbacks[0].assert_called_once_with('job_name')
        generated_callbacks[1].assert_called_once_with(manager, 'job_name', 'event_name')
        expected_calls = [
            mock.call('src1', path('/var/vcap/jobs/job_name/dest1'), 'map', spec,
                      templates_dir=path('/var/vcap/jobs/job_name/templates')),
            mock.call('src2', path('/var/vcap/jobs/job_name/dest2'), 'map', spec,
                      templates_dir=path('/var/vcap/jobs/job_name/templates')),
            mock.call('monit', path('/etc/monit/conf.d/job_name'), 'map', spec,
                      templates_dir=path('/var/vcap/jobs/job_name')),
        ]
        for expected_call in expected_calls:
            self.assertIn(expected_call, RubyTemplateCallback.call_args_list)
        self.assertEqual(RubyTemplateCallback.call_count, len(expected_calls))
        load_spec.assert_called_once_with('job_name')

    @mock.patch('cloudfoundry.tasks.open', create=True)
    @mock.patch('subprocess.check_call')
    @mock.patch('charmhelpers.core.host.chdir')
    def test_patch_dea(self, chdir, call, mopen):
        mopen.return_value.read.return_value = ''
        tasks.patch_dea('service')
        call.assert_called_once_with(['patch', '-s', '-F4'], stdin=mock.ANY)

    def test_enable_swapaccounting(self):
        sample = '''
#comment
text
kernel /boot/vmlinuz/3.x.y arg1

kernel /boot/vmlinuz.3.x.y recovery arg2
        '''
        output = tasks._enable_swapaccounting(sample)
        self.assertIn("arg1 cgroup_enable=memory swapaccount=1", output)
        self.assertIn("recovery arg2 cgroup_enable=memory swapaccount=1", output)
